"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.evaluateExpression = void 0;
var traverse_1 = __importDefault(require("@babel/traverse"));
var t = __importStar(require("@babel/types"));
var ast_1 = require("../ast");
var is_compiled_1 = require("../is-compiled");
var resolve_binding_1 = require("../resolve-binding");
var common_1 = require("./common");
var traverse_access_path_1 = require("./traverse-access-path");
/**
 * Will look in an expression and return the actual value along with updated metadata.
 *
 * E.g: If there is an identifier called `color` that is set somewhere as `const color = 'blue'`,
 * passing the `color` identifier to this function would return `'blue'`.
 *
 * @param expression Expression we want to interrogate.
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var traverseIdentifier = function (expression, meta) {
    var _a;
    var value = undefined;
    var updatedMeta = meta;
    var resolvedBinding = (0, resolve_binding_1.resolveBinding)(expression.name, updatedMeta);
    if (resolvedBinding && resolvedBinding.constant && resolvedBinding.node) {
        // We recursively call get interpolation until it not longer returns an identifier or member expression
        (_a = (0, exports.evaluateExpression)(resolvedBinding.node, resolvedBinding.meta), value = _a.value, updatedMeta = _a.meta);
    }
    return (0, common_1.createResultPair)(value, updatedMeta);
};
/**
 * Returns the binding identifier for a member expression.
 *
 * For example:
 * 1. Member expression `foo.bar.baz` will return the `foo` identifier along
 * with `originalBindingType` as 'Identifier'.
 * 2. Member expression with function call `foo().bar.baz` will return the
 * `foo` identifier along with `originalBindingType` as 'CallExpression'.
 * 3. We also want to process call expressions with a member expression callee
 * i.e. `foo.bar.baz()`
 * @param expression - Member expression node.
 */
var getMemberExpressionMeta = function (expression) {
    var accessPath = [];
    var bindingIdentifier = null;
    var originalBindingType = 'Identifier';
    (0, traverse_1.default)(t.expressionStatement(expression), {
        noScope: true,
        MemberExpression: function (path) {
            // Skip if member comes from call expression arguments
            if (path.listKey === 'arguments') {
                return;
            }
            if (t.isIdentifier(path.node.object)) {
                bindingIdentifier = path.node.object;
                originalBindingType = bindingIdentifier.type;
            }
            else if (t.isCallExpression(path.node.object)) {
                if (t.isIdentifier(path.node.object.callee)) {
                    bindingIdentifier = path.node.object.callee;
                    originalBindingType = path.node.object.type;
                }
            }
            if (t.isIdentifier(path.node.property)) {
                accessPath.push(path.node.property);
            }
            else if (
            // Adds the function name of the trailing call expression
            t.isCallExpression(path.node.property) &&
                t.isIdentifier(path.node.property.callee)) {
                accessPath.push(path.node.property.callee);
            }
        },
    });
    return {
        accessPath: accessPath.reverse(),
        bindingIdentifier: bindingIdentifier,
        originalBindingType: originalBindingType,
    };
};
/**
 * Will look in an expression and return the actual value along with updated metadata.
 *
 * E.g: If there is a member expression called `colors.primary` that has identifier `color` which
 * is set somewhere as `const colors = { primary: 'blue' }`,
 * passing the `colors` identifier to this function would return `'blue'`.
 *
 * @param expression Expression we want to interrogate.
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var traverseMemberExpression = function (expression, meta) {
    var _a = getMemberExpressionMeta(expression), accessPath = _a.accessPath, bindingIdentifier = _a.bindingIdentifier;
    if (bindingIdentifier) {
        return (0, traverse_access_path_1.traverseMemberAccessPath)(t.identifier(bindingIdentifier.name), meta, bindingIdentifier.name, accessPath, expression, { callExpression: traverseCallExpression, memberExpression: traverseMemberExpression });
    }
    return (0, common_1.createResultPair)(expression, meta);
};
/**
 * Will look in an expression and return the actual value along with updated metadata.
 *
 * E.g: If there was a function called `size` that is set somewhere as
 * `const size = () => 10` or `const size = function() { return 10; }` or `function size() { return 10; }`,
 * passing the `size` identifier to this function would return `10` (it will recursively evaluate).
 *
 * @param expression Expression we want to interrogate.
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var traverseFunction = function (expression, meta) {
    var _a;
    var value = undefined;
    var updatedMeta = meta;
    if (t.isBlockStatement(expression.body)) {
        (0, traverse_1.default)(expression.body, {
            noScope: true,
            ReturnStatement: function (path) {
                var _a;
                var argument = path.node.argument;
                if (argument) {
                    (_a = (0, exports.evaluateExpression)(argument, meta), value = _a.value, updatedMeta = _a.meta);
                }
                path.stop();
            },
        });
    }
    else {
        (_a = (0, exports.evaluateExpression)(expression.body, meta), value = _a.value, updatedMeta = _a.meta);
    }
    return (0, common_1.createResultPair)(value, updatedMeta);
};
/**
 * Will find the function node for the call expression and wrap an IIFE around it (to avoid name collision)
 * and move all the parameters mapped to passed arguments in the IIFE's scope (own scope in this case).
 * It will also set own scope path so that when we recursively evaluate any node,
 * we will look for its binding in own scope first, then parent scope.
 *
 * @param expression Expression we want to interrogate.
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var traverseCallExpression = function (expression, meta) {
    var _a;
    var callee = expression.callee;
    var value = undefined;
    // Make sure updatedMeta is a new object, so that when the ownPath is set, the meta does not get re-used incorrectly in
    // later parts of the AST
    var updatedMeta = __assign({}, meta);
    /*
      Basically flow is as follows:
      1. Get the calling function node. It can be either identifier `func('arg')` or a
         member expression `x.func('arg')`;
      2. Save the reference of function node for example `func`.
      3. Pull out its parameters.
      4. Evaluate the args of calling function `func('arg')` by traversing recursively.
      5. Loop through params, and map the args with params.
      6. Pull the params and create an IIFE around the calling function `CallExpression` which creates
         a new scope around calling function and isolates everything (We add bindings at this step by pushing into its scope).
      7. Add that IIFE node path to `ownPath`, and check only for own binding in `resolveBindingNode` function
         so that only isolated params are evaluated.
      8. In `resolveBindingNode`, if `ownPath` is not set (module traversal case or any other case), it will pick things from `parentPath`.
    */
    if (t.isExpression(callee)) {
        var functionNode = void 0;
        if (t.isFunction(callee)) {
            functionNode = callee;
        }
        else {
            // Get func node either from `Identifier` i.e. `func('arg')` or `MemberExpression` i.e. `x.func('arg')`.
            // Right now we are only supported these 2 flavors. If we have complex case like `func('arg').fn().variable`,
            // it will not get evaluated.
            if (t.isIdentifier(callee)) {
                var resolvedBinding = (0, resolve_binding_1.resolveBinding)(callee.name, updatedMeta);
                if (resolvedBinding && resolvedBinding.constant) {
                    functionNode = resolvedBinding.node;
                }
            }
            else if (t.isMemberExpression(callee) && t.isIdentifier(callee.property)) {
                // Convert the last property of the member to a call expression so we attach
                // the functions arguments
                callee.property = t.callExpression(callee.property, expression.arguments);
                return traverseMemberExpression(callee, updatedMeta);
            }
        }
        // Check if it is resolved
        if (functionNode && t.isFunction(functionNode)) {
            // Pull its parameters
            var params = functionNode.params;
            // Evaluate the passed args recursively
            var evaluatedArguments_1 = expression.arguments.map(function (argument) { return (0, exports.evaluateExpression)(argument, updatedMeta).value; });
            // Get path of the call expression `func('arg')` or `x.func('arg')`
            var expressionPath = (0, ast_1.getPathOfNode)(expression, updatedMeta.parentPath);
            // Create an IIFE around it and replace its path. So `func('arg')` will become `(() => func('arg'))()`
            // and `x.func('arg')` will become `(() => x.func('arg'))()`.
            // `wrappingNodePath` is the path above.
            var wrappingNodePath = expressionPath.replaceWith((0, ast_1.wrapNodeInIIFE)(expression))[0];
            // Get arrowFunctionExpressionPath, which was created using the IIFE
            var arrowFunctionExpressionPath_1 = (0, ast_1.getPathOfNode)(wrappingNodePath.node.callee, wrappingNodePath);
            // Loop through the parameters. Right now only identifier `param` and object pattern `{ param }` or `{ param: p }`
            // are supported.
            params
                .filter(function (param) { return t.isIdentifier(param) || t.isObjectPattern(param); })
                .forEach(function (param, index) {
                var evaluatedArgument = evaluatedArguments_1[index];
                // Push evaluated args and params in the IIFE's scope by created a local variable
                // `const param = 'evaluated arg value'`
                arrowFunctionExpressionPath_1.scope.push({
                    id: param,
                    init: evaluatedArgument,
                    kind: 'const',
                });
            });
            // Set the `ownPath` which `resolveBindingNode` will use to check own binding for the
            // local variables we created above.
            updatedMeta.ownPath = arrowFunctionExpressionPath_1;
        }
        (_a = (0, exports.evaluateExpression)(callee, updatedMeta), value = _a.value, updatedMeta = _a.meta);
    }
    return (0, common_1.createResultPair)(value, updatedMeta);
};
/**
 * Returns `true` if an identifier or any paths that reference the identifier are mutated.
 * @param path
 */
var isIdentifierReferencesMutated = function (path) {
    var binding = path.scope.getBinding(path.node.name);
    if (!binding) {
        return false;
    }
    if (!t.isVariableDeclarator(binding.path.node) || !binding.constant) {
        return true;
    }
    for (var i = 0; i < binding.referencePaths.length; i++) {
        var refPath = binding.referencePaths[i];
        var innerBinding = refPath.scope.getBinding(path.node.name);
        if (!innerBinding) {
            continue;
        }
        if (!t.isVariableDeclarator(innerBinding.path.node) || !innerBinding.constant) {
            return true;
        }
    }
    return false;
};
/**
 * Will traverse a path and its identifiers to find all bindings.
 * If any of those bindings are mutated `true` will be returned.
 *
 * @param path
 */
var isPathReferencingAnyMutatedIdentifiers = function (path) {
    if (path.isIdentifier()) {
        return isIdentifierReferencesMutated(path);
    }
    var mutated = false;
    path.traverse({
        Identifier: function (innerPath) {
            var result = isIdentifierReferencesMutated(path);
            if (result) {
                mutated = true;
                // No need to keep traversing - let's stop!
                innerPath.stop();
            }
        },
    });
    return mutated;
};
/**
 * Will try to statically evaluate the node.
 * If successful it will return a literal node,
 * else it will return the fallback node.
 *
 * @param node Node to evaluate
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 * @param fallbackNode Optional node to return if evaluation is not successful. Defaults to `node`.
 */
var babelEvaluateExpression = function (node, meta, fallbackNode) {
    if (fallbackNode === void 0) { fallbackNode = node; }
    var path = (0, ast_1.getPathOfNode)(node, meta.parentPath);
    if (isPathReferencingAnyMutatedIdentifiers(path)) {
        return fallbackNode;
    }
    var result = path.evaluate();
    if (result.value) {
        switch (typeof result.value) {
            case 'string':
                return t.stringLiteral(result.value);
            case 'number':
                return t.numericLiteral(result.value);
        }
    }
    return fallbackNode;
};
/**
 * Will look in an expression and return the actual value along with updated metadata.
 * If the expression is an identifier node (a variable) and a constant,
 * it will return the variable reference.
 *
 * E.g: If there was a identifier called `color` that is set somewhere as `const color = 'blue'`,
 * passing the `color` identifier to this function would return `'blue'`.
 *
 * This behaviour is the same for const string & numeric literals,
 * and object expressions.
 *
 * @param expression Expression we want to interrogate.
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var evaluateExpression = function (expression, meta) {
    var _a, _b, _c, _d;
    var value = undefined;
    var updatedMeta = meta;
    // --------------
    // NOTE: We are recursively calling evaluateExpression() which is then going to try and evaluate it
    // multiple times. This may or may not be a performance problem - when looking for quick wins perhaps
    // there is something we could do better here.
    // --------------
    if (t.isIdentifier(expression)) {
        (_a = traverseIdentifier(expression, updatedMeta), value = _a.value, updatedMeta = _a.meta);
    }
    else if (t.isMemberExpression(expression)) {
        (_b = traverseMemberExpression(expression, updatedMeta), value = _b.value, updatedMeta = _b.meta);
    }
    else if (t.isFunction(expression)) {
        (_c = traverseFunction(expression, updatedMeta), value = _c.value, updatedMeta = _c.meta);
    }
    else if (t.isCallExpression(expression)) {
        (_d = traverseCallExpression(expression, updatedMeta), value = _d.value, updatedMeta = _d.meta);
    }
    if (t.isStringLiteral(value) ||
        t.isNumericLiteral(value) ||
        t.isObjectExpression(value) ||
        t.isTaggedTemplateExpression(value) ||
        // TODO this should be more generic
        (value && (0, is_compiled_1.isCompiledKeyframesCallExpression)(value, updatedMeta.state))) {
        return (0, common_1.createResultPair)(value, updatedMeta);
    }
    if (value) {
        // If we fail to statically evaluate `value` we will return `expression` instead.
        // It's preferable to use the identifier than its result if it can't be statically evaluated.
        // E.g. say we got the result of an identifier `foo` as `bar()` -- its more preferable to return
        // `foo` instead of `bar()` for a single source of truth.
        var babelEvaluatedNode_1 = babelEvaluateExpression(value, updatedMeta, expression);
        return (0, common_1.createResultPair)(babelEvaluatedNode_1, updatedMeta);
    }
    var babelEvaluatedNode = babelEvaluateExpression(expression, updatedMeta);
    return (0, common_1.createResultPair)(babelEvaluatedNode, updatedMeta);
};
exports.evaluateExpression = evaluateExpression;
//# sourceMappingURL=index.js.map