"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.resolveBinding = exports.resolveIdentifierComingFromDestructuring = void 0;
var fs_1 = __importDefault(require("fs"));
var path_1 = require("path");
var parser_1 = require("@babel/parser");
var traverse_1 = __importDefault(require("@babel/traverse"));
var t = __importStar(require("@babel/types"));
var resolve_1 = __importDefault(require("resolve"));
var constants_1 = require("../constants");
var traversers_1 = require("./traversers");
/**
 * Will recursively checks if identifier name is coming from destructuring. If yes,
 * then will return the resolved identifier. We can look for identifier name
 * either in destructuring key or its value.
 *
 * @param name Identifier name to resolve
 * @param node Any Expression node
 * @param resolveFor Either resolve destructuring key or its value
 */
var resolveIdentifierComingFromDestructuring = function (_a) {
    var name = _a.name, node = _a.node, _b = _a.resolveFor, resolveFor = _b === void 0 ? 'key' : _b;
    var resolvedDestructuringIdentifier;
    if (t.isObjectPattern(node)) {
        var pattern = node;
        return pattern.properties.find(function (property) {
            if (t.isObjectProperty(property)) {
                if (resolveFor === 'key') {
                    return t.isIdentifier(property.key) && property.key.name === name;
                }
                else if (resolveFor === 'value') {
                    return t.isIdentifier(property.value) && property.value.name === name;
                }
            }
            return false;
        });
    }
    else if (t.isVariableDeclarator(node)) {
        var declarator = node;
        resolvedDestructuringIdentifier = (0, exports.resolveIdentifierComingFromDestructuring)({
            name: name,
            node: declarator.id,
            resolveFor: resolveFor,
        });
    }
    return resolvedDestructuringIdentifier;
};
exports.resolveIdentifierComingFromDestructuring = resolveIdentifierComingFromDestructuring;
/**
 * Will resolve the value `node` for identifier present inside destructuring
 * If value `node` resolves to an identifier, it will recursively search for its
 * value `node`.
 *
 * For eg.
 * 1. If there is an identifier `foo`, coming from destructuring `{ bar: foo }`
 * having value node as `{ bar: 10 }`, it will resolve to `NumericalLiteral` node `10`.
 * 2. If there is an identifier `foo`, coming from destructuring `{ baz: foo }`
 * referencing an identifier `bar` which in turn having value node as `{ baz: 10 }`,
 * it will search recursively and resolve to `NumericalLiteral` node `10`.
 *
 * @param expression Node inside which we have to resolve the value
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 * @param referenceName {string} Reference name for which `binding` to be resolved
 */
var resolveObjectPatternValueNode = function (expression, meta, referenceName) {
    var objectPatternValueNode = undefined;
    if (t.isObjectExpression(expression)) {
        (0, traverse_1.default)(expression, {
            noScope: true,
            ObjectProperty: {
                exit: function (path) {
                    if (t.isIdentifier(path.node.key, { name: referenceName })) {
                        objectPatternValueNode = path.node.value;
                        path.stop();
                    }
                },
            },
        });
    }
    else if (t.isIdentifier(expression)) {
        var resolvedBinding = (0, exports.resolveBinding)(expression.name, meta);
        if (resolvedBinding) {
            var isResolvedToSameNode = resolvedBinding.path.node === expression;
            if (!isResolvedToSameNode &&
                resolvedBinding.constant &&
                t.isExpression(resolvedBinding.node)) {
                objectPatternValueNode = resolveObjectPatternValueNode(resolvedBinding.node, meta, referenceName);
            }
        }
    }
    return objectPatternValueNode;
};
/**
 * Will return property key if it is different from its value and its value is
 * equal to reference name for which we have to resolve binding.
 *
 * Eg. If we have something like `const { key: value } = { key: 'something' }`, and
 * reference name is `value`, it will return `key` so that it can be resolved to
 * 'something' otherwise it won't get resolved.
 * Input: `node: const { key: value } = { key: 'something' }, referenceName: 'value'`
 * Output: `'key'`
 *
 * Input: `node: const { key } = { key: 'something' }, referenceName: 'value'`
 * Output: `'value'`
 *
 * Input: `node: const { key } = { key: 'something' }, referenceName: 'key'`
 * Output: `'key'`

 * @param node Object pattern node which we have to investigate
 * @param referenceName Reference name for which `binding` to be resolved
 */
var getDestructuredObjectPatternKey = function (node, referenceName) {
    var result = referenceName;
    for (var _i = 0, _a = node.properties; _i < _a.length; _i++) {
        var property = _a[_i];
        if (t.isObjectProperty(property)) {
            var keyName = t.isIdentifier(property.key) ? property.key.name : '';
            var keyValue = t.isIdentifier(property.value) ? property.value.name : '';
            if (keyName !== keyValue && keyValue === referenceName) {
                result = keyName;
                break;
            }
        }
    }
    return result;
};
var resolveRequest = function (request, extensions, meta) {
    var _a = meta.state, filename = _a.filename, opts = _a.opts;
    var resolver = opts.resolver;
    if (!filename) {
        throw new Error('Unable to resolve request due to a missing filename, this is probably a bug!');
    }
    if (!resolver) {
        var id = request.charAt(0) === '.' ? (0, path_1.join)((0, path_1.dirname)(filename), request) : request;
        return resolve_1.default.sync(id, {
            extensions: extensions,
        });
    }
    return resolver.resolveSync(filename, request);
};
/**
 * Will return the `node` of the a binding.
 * This function will follow import specifiers to return the actual `node`.
 *
 * When wanting to do further traversal on the resulting `node` make sure to use the output `meta` as well.
 * The `meta` will be for the resulting file it was found in.
 *
 * @param referenceName {string} Reference name for which `binding` to be resolved
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var resolveBinding = function (referenceName, meta) {
    var _a, _b;
    var _c, _d, _e;
    // Check binding in own scope first so that manually created scopes can be
    // evaluated first then parent scopes or scopes coming from different module.
    var binding = ((_c = meta.ownPath) === null || _c === void 0 ? void 0 : _c.scope.getOwnBinding(referenceName)) ||
        meta.parentPath.scope.getBinding(referenceName);
    if (!binding || binding.path.isObjectPattern()) {
        // Bail early if there is no binding or its a node that we don't want to resolve
        // such as an destructured args from a function.
        return undefined;
    }
    if (t.isVariableDeclarator(binding.path.node)) {
        var node = binding.path.node.init;
        if (t.isObjectPattern(binding.path.node.id) && t.isExpression(node)) {
            node = resolveObjectPatternValueNode(node, meta, getDestructuredObjectPatternKey(binding.path.node.id, referenceName));
        }
        return {
            meta: meta,
            node: node,
            path: binding.path,
            constant: binding.constant,
            source: 'module',
        };
    }
    if ((_d = binding.path.parentPath) === null || _d === void 0 ? void 0 : _d.isImportDeclaration()) {
        // NOTE: We're skipping traversal when file name is not resolved. Imported identifier
        // will end up as a dynamic variable instead.
        if (!meta.state.filename) {
            return;
        }
        var moduleImportSource = binding.path.parentPath.node.source.value;
        // Babel Plugin cannot differentiate between a variable and reserved keywords (e.g keyframes)
        // It will therefore try to parse and resolve both.
        // This workaround shortcircuits when we call `resolveBinding` on a Compiled module.
        // Documented in Issue ##1010: https://github.com/atlassian-labs/compiled/issues/1010
        if (moduleImportSource.startsWith('@compiled/')) {
            // Ignore @compiled modules.
            return;
        }
        var extensions = (_e = meta.state.opts.extensions) !== null && _e !== void 0 ? _e : constants_1.DEFAULT_CODE_EXTENSIONS;
        var modulePath_1 = resolveRequest(moduleImportSource, extensions, meta);
        if (!extensions.some(function (extension) { return modulePath_1.endsWith(extension); })) {
            // Don't attempt to parse any files that are not configured as code
            return;
        }
        var moduleCode_1 = meta.state.cache.load({
            namespace: 'read-file',
            cacheKey: modulePath_1,
            value: function () { return fs_1.default.readFileSync(modulePath_1, 'utf-8'); },
        });
        var ast_1 = meta.state.cache.load({
            namespace: 'parse-module',
            cacheKey: modulePath_1,
            value: function () { return (0, parser_1.parse)(moduleCode_1, { sourceType: 'module', sourceFilename: modulePath_1 }); },
        });
        var foundNode = undefined;
        var foundParentPath = undefined;
        if (binding.path.isImportDefaultSpecifier()) {
            (_a = meta.state.cache.load({
                namespace: 'find-default-export-module-node',
                cacheKey: modulePath_1,
                value: function () {
                    var result = (0, traversers_1.getDefaultExport)(ast_1);
                    return {
                        foundNode: result === null || result === void 0 ? void 0 : result.node,
                        foundParentPath: result === null || result === void 0 ? void 0 : result.path,
                    };
                },
            }), foundNode = _a.foundNode, foundParentPath = _a.foundParentPath);
        }
        else if (binding.path.isImportSpecifier()) {
            var imported = binding.path.node.imported;
            var exportName_1 = t.isIdentifier(imported) ? imported.name : imported.value;
            (_b = meta.state.cache.load({
                namespace: 'find-named-export-module-node',
                cacheKey: "modulePath=".concat(modulePath_1, "&exportName=").concat(exportName_1),
                value: function () {
                    var result = (0, traversers_1.getNamedExport)(ast_1, exportName_1);
                    return {
                        foundNode: result === null || result === void 0 ? void 0 : result.node,
                        foundParentPath: result === null || result === void 0 ? void 0 : result.path,
                    };
                },
            }), foundNode = _b.foundNode, foundParentPath = _b.foundParentPath);
        }
        else if (binding.path.isImportNamespaceSpecifier()) {
            // There's no node inside the file to reference for namespace imports
            // i.e. import * as theme from 'theme';
            // Therefore we just return the binding path
            var path = binding.path;
            foundNode = path.node;
            foundParentPath = path.parentPath;
        }
        if (!foundNode || !foundParentPath) {
            return undefined;
        }
        return {
            constant: binding.constant,
            node: foundNode,
            path: foundParentPath,
            source: 'import',
            meta: __assign(__assign({}, meta), { parentPath: foundParentPath, state: __assign(__assign({}, meta.state), { file: ast_1, filename: modulePath_1 }) }),
        };
    }
    return {
        node: binding.path.node,
        path: binding.path,
        constant: binding.constant,
        source: 'module',
        meta: meta,
    };
};
exports.resolveBinding = resolveBinding;
//# sourceMappingURL=resolve-binding.js.map